#!/usr/bin/env python3
"""
Cylindrical mesocrystal on a substrate.
Calculation using exact computation and 
an approximate algorithm of the Fourier sum with a cutoff.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2, R3


def get_sample():
    # Materials
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-05, 2e-08)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Basis particle
    inner_ff = ba.Sphere(3*nm)
    inner_particle = ba.Particle(particle_mat, inner_ff)

    # 3D lattice
    lattice = ba.Lattice3D(R3(8*nm, 0, 0), R3(0, 8*nm, 0),
                           R3(0, 0, 8*nm))

    # Crystal
    crystal = ba.Crystal(inner_particle, lattice)

    # Mesocrystal: shaped crystal
    outer_ff = ba.Cylinder(20*nm, 50*nm)
    outer_particle = ba.Mesocrystal(crystal, outer_ff)

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_1.deposit2D(ba.Dilute2D(0.0001/nm2, outer_particle))
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


<%= include_partial("template/sim/gisas_standard_wide.py") -%>


def simulate(title, use_fourier):
    sample = get_sample()
    simulation = get_simulation(sample)
    if use_fourier:
        simulation.options().setMesoReciprocalSum(True, 2.5)
    result = simulation.simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':
    <%- unless test_mode || figure_mode -%>
    ba.showSample3D(get_sample(), sample_size=250*nm, seed=0)
    <%- end -%>
    results = [
        simulate('exact sum', False),
        simulate('Fourier sum', True),
    ]
<%= include_partial("template/main/plot_row_unitaspect.py") -%>



#!/usr/bin/env python3
"""
An example of using the Bayesian sampling library emcee with BornAgain.

Author: Andrew McCluskey (andrew.mccluskey@ess.eu)
"""

import bornagain as ba, corner, emcee, numpy as np, os, scipy
import matplotlib.pyplot as plt
from bornagain import nm


np.random.seed(1)

datadir = os.getenv('BA_DATA_DIR', '')
if not datadir:
    raise Exception("Environment variable BA_DATA_DIR not set")


def get_sample(ni_thickness, ti_thickness):
    # pure real scattering-length densities (in angstrom^-2)
    si_sld_real = 2.0704e-06  # Si (substrate)
    ni_sld_real = 9.4245e-06  # Ni
    ti_sld_real = -1.9493e-06  # Ti

    # materials
    vacuum = ba.Vacuum()
    ni_color = (0.93, 0.48, 0.14)
    ni_mat = ba.SLDMaterial("Ni", ni_color, ni_sld_real, 0)
    ti_color = (0.05, 0.62, 0.55)
    ti_mat = ba.SLDMaterial("Ti", ti_color, ti_sld_real, 0)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("SiSubstrate", substrate_color, si_sld_real, 0)

    # layers
    vacuum_layer = ba.Layer(vacuum)
    ni_layer = ba.Layer(ni_mat, ni_thickness)
    ti_layer = ba.Layer(ti_mat, ti_thickness)
    substrate_layer = ba.Layer(substrate_mat)

    # periodic stack
    n_repetitions = 10
    stack = ba.LayerStack(n_repetitions)
    stack.addLayer(ti_layer)
    stack.addLayer(ni_layer)

    # sample
    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addStack(stack)
    sample.addLayer(substrate_layer)

    return sample


def get_simulation(sample, points):
    scan = ba.AlphaScan(ba.ListScan("alpha_i (rad)", points))
    scan.setWavelength(0.154 * nm);
    return ba.SpecularSimulation(scan, sample)


def run_simulation(points, ni_thickness, ti_thickness):
    sample = get_sample(ni_thickness, ti_thickness)
    simulation = get_simulation(sample, points)

    result = simulation.simulate()
    return result.intensities()



if __name__ == '__main__':
    filepath = os.path.join(datadir, "specular/genx_alternating_layers.dat.gz")
    flags = ba.ImportSettings1D("2alpha (deg)", "#", "", 1, 2)
    data = ba.readData1D(filepath, ba.csv1D, flags)

    q = data.xCenters()
    y = data.intensities()
    dy = y * 0.1 # arbitrary uncertainties

    def log_likelihood(P):
        """
        Calculates the log-likelihood for the normal uncertainties
        :tuple sim_var: the variable parameters
        :array x: the abscissa data (q-values)
        :array y: the ordinate data (R-values)
        :array yerr: the ordinate uncertainty (dR-values)
        :return: log-likelihood
        """
        y_sim = run_simulation(q, *P)
        sigma2 = dy**2 + y_sim**2
        return -0.5*np.sum((y - y_sim)**2/sigma2 + np.log(sigma2))

    # Use differential evolution to find the maximum likelihood estimate
    objective_function = lambda *args: -log_likelihood(*args) # just change sign
    initial = np.array([9.0, 1.0]) + 0.1*np.random.randn(2)
    solution = scipy.optimize.differential_evolution(
        objective_function,
        bounds=((5.0, 9.0), (1.0, 10.0)))

    ni_thickness_ml, ti_thickness_ml = solution.x
    print('MLE Ni Thickness', ni_thickness_ml, 'nm')
    print('MLE Ti Thickness', ti_thickness_ml, 'nm')

    # Perform the likelihood sampling
    nwalkers = 32
    ndim = 2
    pos = solution.x + 1e-4*np.random.randn(nwalkers, ndim)
    sampler = emcee.EnsembleSampler(nwalkers,
                                    ndim,
                                    log_likelihood)
    sampler.run_mcmc(pos, 1, progress=True)

    # Plot and show corner plot of samples
    flat_samples = sampler.get_chain(flat=True)
    corner.corner(flat_samples,
                  labels=['Ni-thickness/nm', 'Ti-thickness/nm'])


    # Plot and show MLE and data of reflectivity
    plt.errorbar(q, y, dy, marker='.', ls='')
    plt.plot(
        q,
        run_simulation(q, *flat_samples.mean(axis=0)),
        '-')
    plt.xlabel('$\\alpha$/rad')
    plt.ylabel('$R$')
    plt.yscale('log')



#!/usr/bin/env python3
"""
Dump formfactor for different q scans
"""

import math
import bornagain as ba

eps = 0.005

if __name__ == '__main__':

    width = 1.
    height = 1.
    length = 100.0

    particle = ba.FormFactorCosineRippleBox(length, width, height)
#    particle = ba.FormFactorSawtoothRippleBox(length, width, height, 0.)

    print("# " + particle.shapeName() + " form factor, for different q scans")

    print("# q vs |F(q)| for q in direction 010")
    t = 0.2
    while t < 200:
        q = ba.C3(0, t, 0)
        print(t, " ", abs(particle.formfactor_at_bottom(q)))
        t *= 1+eps
    print()

    print("# q vs |F(q)| for q in direction 0001")
    t = 0.2
    while t < 200:
        q = ba.C3(0, 0, t)
        print(t, " ", abs(particle.formfactor_at_bottom(q)))
        t *= 1+eps
    print()

    print("# q vs |F(q)| for q in direction 011")
    t = 0.2
    s2 = math.sqrt(2)
    while t < 200:
        q = ba.C3(0, t/s2, t/s2)
        print(t, " ", abs(particle.formfactor_at_bottom(q)))
        t *= 1+eps
    print()

    print("# q vs |F(q)| for |q|=50, q on cercle in 100 plane")
    ampl = 50
    t = 0
    while t < 2*math.pi:
        q = ba.C3(0, ampl*math.cos(t), ampl*math.sin(t))
        print(t, " ", abs(particle.formfactor_at_bottom(q)))
        t += eps
    print()



#!/usr/bin/env python

purpose = '''
Convert Python script from BornAgain-21 to BornAgain-22.

Some conversions are too complicated to be automatized.
If unsupported constructs are encountered, this script will fail.

If a converted scripts still does not work under BornAgain-22,
then this is a bug, and should be reported to the maintainers.
'''

import os, re, sys

def replace_t(t):

    substitutions = {
        r'ba\.MultiLayer\(': 'ba.Sample(',
        r'setAngleDistribution\(': 'setGrazingAngleDistribution(',
        r'\n(\s*)bp.plot_simulation_result(\(.+?\))\s*\n': r'\n\1bp.plot_datafield\2\n\1bp.plt.show()\n',
        r'\n(\s*)bp.plot_multicurve_specular(\(.+?\))\s*\n': r'\n\1bp.plot_multicurve\2\n\1bp.plt.show()\n',
        r'\n(\s*)bp.make_plot(\(.+?\))\s*\n': r'\n\1bp.plot_to_grid\2\n\1bp.plt.show()\n',
        r'\n(\s*)bp.make_plot_row(\(.+?\))\s*\n': r'\n\1bp.plot_to_row\2\n\1bp.plt.show()\n',
        r'\n(\s*)bp.plot_histogram(\(.+?\))\s*\n+\s*[^\n]*?show_or_export\(\)\s*\n': r'\n\1bp.plot_datafield\2\n\1bp.plt.show()\n',
        r'\n\s*[^\n]*?parse_args\((.+?)\)\s*\n(.*bp.plot_(datafield|multicurve|to_grid|to_row)\(.*?)\)': r'\n\2, \1)',
    }
    for old, new in substitutions.items():
        t = re.sub(old, new, t, count=0, flags=re.DOTALL)

    prohibited = {
        r'plot_simulation_result',
        r'plot_multicurve_specular',
        r'show_or_export',
        r'parse_args',
        r'RectangularDetector',
        r'Roughness',
        r'RoughnessModel',
        r'CoreShell',
        r'Spheroid',
        r'Truncated',
        r'Coords_',
        r'setCrossCorrLength',
        r'bp\.datfield',
        r'result\.array',
        r'addSimulationAndData',
    }
    for old in prohibited:
        if len(re.findall(old, t)) > 0:
            raise Exception(f"Not able to convert '{old}'")

    return t

def main():
    if len(sys.argv) != 3:
        print(f"Usage: python {sys.argv[0]} <old_fname> <new_fname|new_dir>")
        print(purpose)
        sys.exit(1)

    old_fname, new_path = sys.argv[1], sys.argv[2]

    try:
        with open(old_fname, 'r', encoding='utf-8') as f:
            t = f.read()

        new_t = replace_t(t)

        if os.path.isdir(new_path):
            new_fname = os.path.join(new_path, os.path.basename(old_fname))
        else:
            new_fname = new_path

        with open(new_fname, 'w', encoding='utf-8') as f:
            f.write(new_t)

        # copy permissions
        os.chmod(new_fname, os.stat(old_fname).st_mode)

        if new_t == t:
            print(f"{os.path.basename(old_fname)}: OK (nothing changed). Output saved to {new_path}")
        else:
            print(f"{os.path.basename(old_fname)}: OK (substitutions done). Output saved to {new_path}")

    except FileNotFoundError:
        print(f"ERROR: The file '{old_fname}' does not exist.")
        sys.exit(1)
    except Exception as e:
        print(f"{os.path.basename(old_fname)}: ERROR: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Cylinders forming an infinite 2D square lattice
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    """
    A sample with cylinders on a substrate, forming an infinite 2D square lattice
    """

    # Materials
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 0.0006, 2e-08)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Particles
    ff = ba.Cylinder(3*nm, 3*nm)
    particle = ba.Particle(particle_mat, ff)

    # 2D lattice
    lattice = ba.BasicLattice2D(25*nm, 25*nm, 90*deg, 30*deg)
    layout = ba.Crystal2D(particle, lattice)
    profile = ba.Profile2DCauchy(1000*nm, 1000*nm, 0*deg)
    layout.setDecayFunction(profile)

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_1.deposit2D(layout)
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


<%= include_partial("template/sim/gisas_standard_wide.py") -%>


<%= include_partial("template/main/runsim_unitaspect.py",
    "ba.showSample3D(sample, sample_size=300*nm, seed=0)") -%>



#  **************************************************************************  #
"""
#   BornAgain: simulate and fit reflection and scattering
#
#   @file      Wrap/Python/ba_fitmonitor.py
#   @brief     Plotter classes for monitoring fit progress.
#
#   @homepage  http://apps.jcns.fz-juelich.de/BornAgain
#   @license   GNU General Public License v3 or higher (see COPYING)
#   @copyright Forschungszentrum Juelich GmbH 2019
#   @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
"""
#  **************************************************************************  #

import time

from bornagain import ba_plot as bp

try:  # workaround for build servers
    import numpy as np
    from matplotlib import pyplot as plt
    from matplotlib import gridspec
except Exception as e:
    print("In ba_fitmonitor.py: {:s}".format(str(e)))


class Printer:
    """
    Prints fit iteration progress to standard output.
    Replaces the C++ FitPrintService. Use via fit_objective.initPrint(every_nth).
    """

    def __init__(self):
        self._t_start = None
        self._t_last = None

    def __call__(self, fit_objective):
        info = fit_objective.iterationInfo()
        n = info.iterationCount()
        now = time.time()
        if self._t_start is None:
            self._t_start = now
            self._t_last = now
        dt = now - self._t_last
        self._t_last = now
        print(f"Fit iteration {n} Chi2 {info.chi2():.8e} dt {dt:.2f}")
        vals = " ".join(f"{v:.6e}" for v in info.parameterMap().values())
        print(f"P {vals}")


class Plotter:
    """
    Draws fit progress. Base class for simulation-specific classes (PlotterGISAS etc).
    """

    def __init__(self, zmin=None, zmax=None, xlabel=None, ylabel=None):
        self._fig = plt.figure(figsize=(10.25, 7.69))
        self._fig.canvas.draw()
        self._zmin = zmin
        self._zmax = zmax
        self._xlabel = xlabel
        self._ylabel = ylabel

    def __call__(self, fit_objective):
        self.plot(fit_objective)

    def reset(self):
        self._fig.clf()

    def plot(self):
        self._fig.tight_layout()
        plt.pause(0.3)


class PlotterGISAS(Plotter):
    """
    Draws fit progress, for GISAS simulation.
    """

    def __init__(self, zmin=None, zmax=None, xlabel=None, ylabel=None):
        Plotter.__init__(self, zmin, zmax, xlabel, ylabel)

    @staticmethod
    def make_subplot(nplot):
        plt.subplot(2, 2, nplot)
        plt.subplots_adjust(wspace=0.2, hspace=0.2)

    def plot(self, fit_objective):
        Plotter.reset(self)

        data = fit_objective.experimentalData()
        sim_data = fit_objective.simulationResult()
        diff = fit_objective.absoluteDifference()

        self.make_subplot(1)

        # same limits for both plots
        arr = data.intensities()
        zmax = np.amax(arr) if self._zmax is None else self._zmax
        zmin = zmax*1e-6 if self._zmin is None else self._zmin

        bp.plot_heatmap(data,
                        title="Experimental data",
                        intensity_min=zmin,
                        intensity_max=zmax,
                        xlabel=self._xlabel,
                        ylabel=self._ylabel,
                        zlabel='')

        self.make_subplot(2)
        bp.plot_heatmap(sim_data,
                        title="Simulated data",
                        intensity_min=zmin,
                        intensity_max=zmax,
                        xlabel=self._xlabel,
                        ylabel=self._ylabel,
                        zlabel='')

        self.make_subplot(3)
        bp.plot_heatmap(diff,
                        title="Difference",
                        intensity_min=zmin,
                        intensity_max=zmax,
                        xlabel=self._xlabel,
                        ylabel=self._ylabel,
                        zlabel='')

        self.make_subplot(4)
        plt.title('Parameters')
        plt.axis('off')

        iteration_info = fit_objective.iterationInfo()

        plt.text(
            0.01, 0.85, "Iterations  " +
            '{:d}'.format(iteration_info.iterationCount()))
        plt.text(0.01, 0.75,
                 "Chi2       " + '{:8.4f}'.format(iteration_info.chi2()))
        index = 0
        params = iteration_info.parameterMap()
        for key in params:
            plt.text(0.01, 0.55 - index*0.1,
                     '{:30.30s}: {:6.3f}'.format(key, params[key]))
            index = index + 1

        Plotter.plot(self)


class PlotterSpecular:
    """
    Draws fit progress, for specular simulation.
    """

    def __init__(self, pause=0.0):
        self.pause = pause
        self._fig = plt.figure(figsize=(10, 7))
        self._fig.canvas.draw()

    def __call__(self, fit_objective):
        self.plot(fit_objective)

    def plot(self, fit_objective):
        self._fig.clf()
        # retrieving data from fit suite
        exp_data = fit_objective.experimentalData().plottableField()
        sim_data = fit_objective.simulationResult().plottableField()

        # data values
        sim_values = sim_data.intensities()
        exp_values = exp_data.intensities()
        unc_values = exp_data.errors()

        # default font properties dictionary to use
        font = { 'size': 16 }

        plt.yscale('log')
        plt.ylim((0.5*np.min(exp_values), 5*np.max(exp_values)))
        plt.plot(exp_data.xCenters(), exp_values, 'k--')
        if unc_values is not None:
            plt.plot(exp_data.xCenters(),
                     exp_values - unc_values,
                     'xkcd:grey',
                     alpha=0.6)
            plt.plot(exp_data.xCenters(),
                     exp_values + unc_values,
                     'xkcd:grey',
                     alpha=0.6)
        plt.plot(sim_data.xCenters(), sim_values, 'b')

        xlabel = bp._get_axes_labels(exp_data)[0]
        xlabel2 = bp._get_axes_labels(sim_data)[0]
        assert xlabel == xlabel2, f'Different labels: "{xlabel}" in exp vs "{xlabel2}" in sim'
        legend = ['Experiment', 'BornAgain']
        if unc_values is not None:
            legend = ['Experiment', r'Exp $\pm \sigma$', 'BornAgain']
        plt.legend(legend, loc='upper right', prop=font)
        plt.xlabel(xlabel, fontdict=font)
        plt.ylabel("Intensity", fontdict=font)

        plotargs = bp.parse_commandline()
        _do_show = plotargs.get('do_show', None)
        if _do_show:
            plt.pause(self.pause)

    def show(self):
        plotargs = bp.parse_commandline()
        _do_show = plotargs.get('do_show', None)
        if _do_show:
            plt.show()



#!/usr/bin/env python3
"""
Basic example of depth-probe simulation:
Computes intensity as function of incident angle alpha and depth z.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm

def get_sample():
    # Materials
    vac_mat = ba.Vacuum()
    a_color = (0.05, 0.62, 0.55)
    a_mat = ba.RefractiveMaterial("A", a_color, 6e-5, 0)
    sub_color = (0.28, 0.57, 0.82)
    sub_mat = ba.RefractiveMaterial("Substrate", sub_color, 3e-05, 0)

    # Sample
    sample = ba.Sample()
    sample.addLayer(ba.Layer(vac_mat))
    sample.addLayer(ba.Layer(a_mat, 100*nm))
    sample.addLayer(ba.Layer(sub_mat))

    return sample


def get_simulation(sample, flags):
    n = 500

    scan = ba.AlphaScan(n, 0*deg, 1*deg)
    scan.setWavelength(0.3*nm)

    z_axis = ba.EquiDivision("z (nm)", n, -130*nm, 30*nm)
    simulation = ba.DepthprobeSimulation(scan, sample, z_axis, flags)
    # BUG The following line should be unnecessary:
    simulation.options().setUseAvgMaterials(False)

    return simulation


def run_example(flags=0):
    sample = get_sample()
    simulation = get_simulation(sample, flags)
    result = simulation.simulate()
    plotargs = bp.parse_commandline()
    bp.plot_datafield(result, intensity_min=1e-12, intensity_max=1e2, **plotargs)
    bp.export(**plotargs)


if __name__ == '__main__':
    run_example()



#!/usr/bin/env python3

"""
Check line length of Python files in or under current working directory.
Used by BornAgain CI to ensure that published example code can be nicely
displayed in the online documentation.

Copyright Forschungszentrum Jülich GmbH 2025.
License:  Public Domain
"""

import glob, os, re, sys
from io import open

if len(sys.argv)<2:
    print( """
Usage: check-line-length <max_mum_chars_per_line>

Returns 1 if any given file in or under the current working directory
    - exceeds the given number of characters per line,
    - or contains a tab.
Returns 0 otherwise.
""" )
    sys.exit(-1)

fworst = ""
lworst = 0
totlin = 0
totexc = 0
tottab = 0
nftot = 0
nfexc = 0
nftab = 0

limit = int(sys.argv[1])
flist = glob.glob(os.path.join('.', "**/*.py"), recursive=True)

for fn in flist:
    # read in
    fd = open( fn, 'r' )
    txt = fd.read().split( '\n' )
    fd.close

    n = len(txt)
    maxlen = 0
    maxwhere = -1
    nexclin = 0
    ntablin = 0
    for i in range(n):
        lin = txt[i]
        if re.match(r'\s*\/\/', lin):
            continue # ignore comment lines
        if '\t' in lin:
            ntablin += 1
            lin = lin.replace( '\t', '    ' )
        l = len(lin)
        if l>maxlen:
            maxlen = l
            maxwhere = i
        if l>limit:
            nexclin += 1

    if nexclin>0:
        print( "%s: %i/%i lines too long, line %i longest with %i chars" %
               (fn, nexclin, n, maxwhere+1, maxlen) )
        totexc += nexclin
        nfexc += 1
    if ntablin>0:
        print( "%s: %i/%i lines contain a TAB" % (fn, ntablin, n) )
        tottab += ntablin
        nftab += 1
    totlin += n
    nftot += 1
    if maxlen>lworst:
        lworst = maxlen
        fworst = fn

exitval = 0
if nfexc>0:
    print( "total: %i/%i lines too long, in %i/%i files, longest in %s with %i chars" %
           (totexc, totlin, nfexc, nftot, fworst, lworst) )
    exitval = 1
if nftab>0:
    print( "total: %i/%i lines contain a TAB, in %i/%i files" %
           (tottab, totlin, nftab, nftot) )
    exitval = 1

sys.exit(exitval)



#!/usr/bin/env python3
"""
GISAS of a thick, highly absorptive Au layer on Si, at extra-long wavelength.

Au (SLD: 3.484e-5, 1.791e-5) on a 10 nm vacuum gap on Si (SLD: 3.842e-7,
6.282e-7) with ErfTransient roughness. Beam wavelength 12 nm tests
the numerics under strong absorption.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    au_color = (0.93, 0.72, 0.25)
    au_mat = ba.SLDMaterial("Au", au_color, 3.48388057043e-05, 1.79057609656e-05)
    si_color = (0.30, 0.62, 0.86)
    si_mat = ba.SLDMaterial("Si", si_color, 3.84197565094e-07, 6.28211531498e-07)

    autocorr = ba.SelfAffineFractalModel(5*nm, 0.5, 10*nm)
    transient = ba.ErfTransient()
    roughness = ba.Roughness(autocorr, transient)

    vacuum_layer = ba.Layer(ba.Vacuum())
    au_layer = ba.Layer(au_mat, 200*nm)
    vacuum_layer_2 = ba.Layer(ba.Vacuum(), 10*nm)
    substrate_layer = ba.Layer(si_mat, roughness)

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(au_layer)
    sample.addLayer(vacuum_layer_2)
    sample.addLayer(substrate_layer)
    return sample


def get_simulation(sample):
    beam = ba.Beam(1e8, 12*nm, 0.2*deg)
    n = <%= test_mode ? 25 : 100 %>
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    simulation.options().setIncludeSpecular(True)
    return simulation


<%= include_partial("template/main/runsim_unitaspect.py",
    "ba.showSample3D(sample, sample_size=120*nm, seed=0)") -%>



#!/usr/bin/env python3
"""
Basic offspec example.
Sample is a grating on a substrate, modelled by a 1d lattice of long boxes
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    # Materials
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-5, 2e-08)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Particles
    box_length = 1000*nm
    ff = ba.Box(box_length, 20*nm, 10*nm)
    particle = ba.Particle(particle_mat, ff)
    particle_rotation = ba.RotationZ(90*deg)
    particle.rotate(particle_rotation)

    # Interference functions
    layout = ba.Crystal1D(particle, 100*nm, 0, 1/box_length)
    profile = ba.Profile1DCauchy(1e6*nm)
    layout.setDecayFunction(profile)

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_1.deposit2D(layout)
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    nla = 573
    naf = 200
    lambda_max = 0.5*nm
    scan = ba.LambdaScan(nla, lambda_max/nla, lambda_max)
    scan.setGrazingAngle(1.5 * deg)
    scan.setIntensity(1e12)
    detector = ba.OffspecDetector(naf, 0*deg, 4*deg, -1*deg, +1*deg)
    simulation = ba.OffspecSimulation(scan, sample, detector)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    simulation.options().setOffspecIntegratorOptions(
        "AdaptiveGaussKronrod",
        {
        "max_subintervals" : 200,
        "abs_tolerance" : 1e-5,
        "rel_tolerance" : 1e-3,
        "GK_rule" : 3
        });
    result = simulation.simulate()
    ba.showSample3D(sample, sample_size=1000*nm, seed=0)
    bp.plot_datafield(result, intensity_min=1)
    bp.plt.show()



#!/usr/bin/env python3
"""
Polarized neutron reflectometry of a 4x Fe/Ni bilayer on Au substrate
with TanhTransient interface roughness.

Two-channel (PP, MM) angle-scan with Y-axis polarization.
The Fe magnetization is aligned along Y; PM and MP are exactly zero.
See PolarizedFeNiBilayerSpinFlipTanh.py for the four-channel variant
(38 deg magnetization tilt, PM and MP non-zero).

Compare PolarizedFeNiBilayerTanhQ.py for the Q-space equivalent.
"""
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm, R3


def get_sample():
    vacuum = ba.Vacuum()
    mag_vec = R3(0, 1e7, 0)
    fe_color = (0.48, 0.32, 0.80)
    fe_mat = ba.SLDMaterial("Fe", fe_color, 8.02e-6, 0, mag_vec)
    ni_color = (0.93, 0.48, 0.14)
    ni_mat = ba.SLDMaterial("Ni", ni_color, 9.4245e-6, 0)
    substrate_color = (0.93, 0.72, 0.25)
    substrate_mat = ba.SLDMaterial("Au", substrate_color, 4.6665e-6, 0)

    autocorr = ba.SelfAffineFractalModel(0.2*nm, 0.7, 25*nm)
    roughness = ba.Roughness(autocorr, ba.TanhTransient())

    stack = ba.LayerStack(4)
    stack.addLayer(ba.Layer(fe_mat, 10*nm, roughness))
    stack.addLayer(ba.Layer(ni_mat, 4*nm, roughness))

    sample = ba.Sample()
    sample.addLayer(ba.Layer(vacuum))
    sample.addStack(stack)
    sample.addLayer(ba.Layer(substrate_mat, roughness))
    return sample


def simulate(sample, pol, an, title):
    n = 500
    scan = ba.AlphaScan(n, 5*deg/n, 5*deg)
    scan.setWavelength(1.54*angstrom)
    scan.setPolarization(pol)
    scan.setAnalyzer(an)
    result = ba.SpecularSimulation(scan, sample).simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':
    sample = get_sample()
    results = [
        simulate(sample, R3(0, +1, 0), R3(0, +1, 0), "PP"),
        simulate(sample, R3(0, -1, 0), R3(0, -1, 0), "MM"),
    ]
    plotargs = bp.parse_commandline()
    bp.plot_multicurve(results, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
Spheres on two hexagonal close packed layers
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3


def get_sample():
    """
    A sample with spheres on a substrate,
    forming two hexagonal close packed layers.
    """

    # Materials
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-7, 2e-08)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Particles
    ff = ba.Sphere(10*nm)
    particle_1 = ba.Particle(particle_mat, ff)
    particle_2 = ba.Particle(particle_mat, ff)
    particle_2_position = R3(10*nm, 10*nm, 17.3205080757*nm)
    particle_2.translate(particle_2_position)

    # Lattice basis as compound particle
    basis = ba.Compound()
    basis.addComponent(particle_1)
    basis.addComponent(particle_2)

    # 2D lattice
    lattice = ba.HexagonalLattice2D(20*nm, 0)
    layout = ba.Crystal2D(basis, lattice)
    profile = ba.Profile2DCauchy(100*nm, 100*nm, 0)
    layout.setDecayFunction(profile)

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_1.deposit2D(layout)
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 200
    detector = ba.SphericalDetector(n, -1*deg, 1*deg, n, 0, 1*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    plotargs = bp.parse_commandline()
    bp.plot_datafield(result, unit_aspect=1, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
Polarized specular reflectivity with B field in -Z direction.

Tests the numerical singularity in eigenToMatrix() when B_z = -1.
Sample: Vacuum | MagLayer (10 nm, B=(0,0,-1e8)) | Substrate.
Two channels: PP and MM with Y-axis polarization.
"""
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm, R3


def get_sample():
    layer_field = R3(0, 0, -1e8)
    layer_color = (0.86, 0.24, 0.18)
    layer_mat = ba.SLDMaterial("MagLayer", layer_color, 1e-4, 1e-8, layer_field)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("Substrate", substrate_color, 7e-5, 2e-6)

    sample = ba.Sample()
    sample.addLayer(ba.Layer(ba.Vacuum()))
    sample.addLayer(ba.Layer(layer_mat, 10*nm))
    sample.addLayer(ba.Layer(substrate_mat))
    return sample


def simulate(sample, pol, an, title):
    n = 500
    scan = ba.AlphaScan(n, 5*deg/n, 5*deg)
    scan.setWavelength(1.54*angstrom)
    scan.setPolarization(pol)
    scan.setAnalyzer(an)
    result = ba.SpecularSimulation(scan, sample).simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':
    sample = get_sample()
    ba.showSample3D(sample, sample_size=80*nm, seed=0)
    results = [
        simulate(sample, R3(0, +1, 0), R3(0, +1, 0), "PP"),
        simulate(sample, R3(0, -1, 0), R3(0, -1, 0), "MM"),
    ]
    bp.plot_multicurve(results)
    bp.plt.show()



#!/usr/bin/env python3

import math
from types import SimpleNamespace

from bornagain.ba_view3d import _particle_matches_mixture_choices, _pick_active_variants


def active_name(seed, site, particles, weights):
    """Returns the single particle name active at a synthetic layout site."""
    choices = _pick_active_variants(seed, 0, float(site), 0.0, 0.0, weights)
    active = [
        name for name, particle in particles.items()
        if _particle_matches_mixture_choices(particle, choices)
    ]
    assert len(active) == 1
    return active[0]


def assert_frequency(name, actual, expected, n):
    """Checks sampled counts against a binomial three-sigma interval."""
    sigma = math.sqrt(n * expected * (1.0 - expected))
    assert abs(actual - n * expected) <= 3.0 * sigma, (
        f"{name}: got {actual}, expected {n * expected} +/- {3.0 * sigma}"
    )


def test_nested_mixture_frequencies():
    # Mixture(A, Mixture(B, C)) with outer weights (0.5, 0.5)
    # and inner weights (0.4, 0.6) gives P(A)=0.5, P(B)=0.2, P(C)=0.3.
    weights = [[0.5, 0.5], [0.4, 0.6]]
    particles = {
        "A": SimpleNamespace(mixture_groups=[0], mixture_variants=[0]),
        "B": SimpleNamespace(mixture_groups=[0, 1], mixture_variants=[1, 0]),
        "C": SimpleNamespace(mixture_groups=[0, 1], mixture_variants=[1, 1]),
    }

    n = 5000
    counts = {"A": 0, "B": 0, "C": 0}
    for site in range(n):
        counts[active_name(123, site, particles, weights)] += 1

    assert_frequency("A", counts["A"], 0.5, n)
    assert_frequency("B", counts["B"], 0.2, n)
    assert_frequency("C", counts["C"], 0.3, n)


def test_mixture_choices_are_deterministic_by_seed_and_position():
    weights = [[0.7, 0.3], [0.2, 0.8]]

    first = [
        _pick_active_variants(17, 2, float(site), 1.25, -3.5, weights)
        for site in range(100)
    ]
    second = [
        _pick_active_variants(17, 2, float(site), 1.25, -3.5, weights)
        for site in range(100)
    ]
    changed_seed = [
        _pick_active_variants(18, 2, float(site), 1.25, -3.5, weights)
        for site in range(100)
    ]

    assert first == second
    assert any(a != b for a, b in zip(first, changed_seed))


if __name__ == "__main__":
    test_nested_mixture_frequencies()
    test_mixture_choices_are_deterministic_by_seed_and_position()



#!/usr/bin/env python3
"""
Example of simulating a reflectometry experiment
with a rough sample using BornAgain.

"""
import bornagain as ba
from bornagain import nm, ba_plot as bp, deg, nm


def get_sample():
    # Materials
    vacuum = ba.Vacuum()
    ti_color = (0.05, 0.62, 0.55)
    ti_mat = ba.SLDMaterial("Ti", ti_color, -1.9493e-06, 0)
    ni_color = (0.93, 0.48, 0.14)
    ni_mat = ba.SLDMaterial("Ni", ni_color, 9.4245e-06, 0)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("SiSubstrate", substrate_color, 2.0704e-06, 0)

    # roughness
    autocorr = ba.SelfAffineFractalModel(1*nm, 0.7, 25*nm)
    transient = ba.TanhTransient()
    roughness = ba.Roughness(autocorr, transient)

    # Layers
    ambient_layer = ba.Layer(vacuum)
    ti_layer = ba.Layer(ti_mat, 3*nm, roughness)
    ni_layer = ba.Layer(ni_mat, 7*nm, roughness)
    substrate_layer = ba.Layer(substrate_mat, roughness)

    # Periodic stack
    n_repetitions = 10
    stack = ba.LayerStack(n_repetitions)
    stack.addLayer(ti_layer)
    stack.addLayer(ni_layer)
    
    # Sample
    sample = ba.Sample()
    sample.addLayer(ambient_layer)
    sample.addStack(stack)
    sample.addLayer(substrate_layer)

    return sample


def get_simulation(sample):
    n = 50
    scan = ba.AlphaScan(n, 2*deg/n, 2*deg)
    scan.setWavelength(0.154*nm)
    return ba.SpecularSimulation(scan, sample)


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#  **************************************************************************  #
"""
#   BornAgain: simulate and fit reflection and scattering
#
#   @file      Wrap/Python/ba_check.py
#   @brief     Infrastructure for persistence tests.
#
#   @homepage  http://apps.jcns.fz-juelich.de/BornAgain
#   @license   GNU General Public License v3 or higher (see COPYING)
#   @copyright Forschungszentrum Juelich GmbH 2016
#   @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
"""
#  **************************************************************************  #

import math, sys
import bornagain as ba

def matches_reference(result, fname, tolerance, reference, subname=""):
    """
    Check simulation result against reference.
    Used internally.
    """

    if not tolerance or not 'reference':
        print(f"{fname}: no tolerance or no reference")
        return True

    reffile = reference + subname + ".int"
    ok = ba.dataMatchesFile(result, reffile, tolerance)
    print(f"{fname} vs {reffile}, tol={tolerance} => ok={ok}")
    return ok

def persistence_test(result):
    for arg in sys.argv[1:]:
        s = arg.split("=")
        if len(s) != 2:
            raise Exception(f"command-line argument '{arg}' does not have form key=value")
        if s[0]=='datfile':
            datfile = s[1]
        elif s[0]=='reference':
            reference = s[1]
        elif s[0]=='tolerance':
            tolerance = float(s[1])
        else:
            raise Exception(f"unexpected argument '{arg}'")

    try:
        datfile
        reference
    except:
        raise Exception("missing some obligatory argument")
    assert(tolerance>0)
    assert(tolerance<1)

    if isinstance(result, list):
        nDigits = int(math.log10(len(result))) + 1
        formatN = "%" + str(nDigits) + "i"

        ok = True
        for i, one_result in enumerate(result):
            ok = ok and matches_reference(one_result, datfile, tolerance, reference, "." + (formatN % i))

        if not ok:
            print("To overwrite references:")
            for i, one_result in enumerate(result):
                outfile = datfile + "." + (formatN % i) + ".int"
                reffile = reference + "." + (formatN % i) + ".int"
                ba.writeDatafield(one_result, outfile)
                print(f"cp -f {outfile} {reffile}")

            raise Exception("No agreement between result and reference")

    else:
        if not matches_reference(result, datfile, tolerance, reference):
            outfile = datfile + ".int"
            reffile = reference + ".int"
            ba.writeDatafield(result, outfile)
            print(f"To overwrite reference:\ncp -f {outfile} {reffile}")
            raise Exception("No agreement between result and reference")



#!/usr/bin/env python3
"""
Spheres in a finite layer. For testing the slicing machinery.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2


def get_sample():
    # defining materials
    vacuum = ba.Vacuum()
    layer1_color = (0.93, 0.72, 0.25)
    layer1_mat = ba.RefractiveMaterial("Layer1", layer1_color, 4e-6, 2e-8)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-6, 2e-8)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 3e-5, 2e-8)

    # particle
    ff = ba.SphericalSegment(5*nm, 0, 5*nm)
    particle = ba.Particle(particle_mat, ff)

    # layers
    layer1 = ba.Layer(layer1_mat, 6*nm)
    layer1.setNumberOfSlices(10)
    layer1.deposit2D(ba.Dense2D(.002/nm2, particle))
    substrate = ba.Layer(substrate_mat)

    sample = ba.Sample()
    sample.addLayer(ba.Layer(vacuum))
    sample.addLayer(layer1)
    sample.addLayer(substrate)
    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 11
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#!/usr/bin/env python3
"""
Polarized GISAS (PP and MP channels) of magnetic spheres in a magnetic
layer.

Spheres with x-oriented magnetisation (1e6 A/m) are deposited on a
magnetic substrate. Two polarisation channels are shown: PP (+Z/+Z, no
spin flip) and MP (-Z/+Z, spin flip from below).
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3


def get_sample():
    vacuum = ba.Vacuum()
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 7e-6, 2e-8)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("MagParticle", particle_color, 6e-4, 2e-8,
                                               R3(1e6, 0, 0))

    ff = ba.Sphere(5*nm)
    particle = ba.Particle(particle_mat, ff)

    vacuum_layer = ba.Layer(vacuum)
    vacuum_layer.deposit2D(ba.Dilute2D(0.001, particle))
    substrate_layer = ba.Layer(substrate_mat)

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(substrate_layer)
    return sample


def get_simulation(sample, pol_dir, an_dir):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 200
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 2*deg)
    beam.setPolarization(pol_dir)
    detector.setAnalyzer(an_dir)
    return ba.ScatteringSimulation(beam, sample, detector)


def simulate(pol_dir, an_dir):
    sample = get_sample()
    simulation = get_simulation(sample, pol_dir, an_dir)
    return simulation.simulate()


if __name__ == '__main__':
    z_up = R3(0, 0, 1)
    z_dn = R3(0, 0, -1)
    results = [
        simulate(z_up, z_up),   # PP
        simulate(z_dn, z_up),   # MP
    ]
    ba.showSample3D(get_sample(), sample_size=120*nm, seed=0)
    bp.plot2d_to_row(results, unit_aspect=1)
    bp.plt.show()



#!/usr/bin/env python3
"""
Spin-flip scattering by magnetic spheres
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2, R3


def get_sample():
    # Materials
    B = R3(0, 0, 1e7)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 2e-05, 4e-07, B)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial(
        "Substrate", substrate_color, 7e-06, 1.8e-07)
    vacuum = ba.Vacuum()

    # Particles
    ff = ba.Sphere(5*nm)
    particle = ba.Particle(particle_mat, ff)

    # Multilayer
    layer_1 = ba.Layer(vacuum)
    layer_2 = ba.Layer(substrate_mat)
    layer_2.suspend2D(ba.Dilute2D(0.01/nm2, particle))

    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e12, 0.1*nm, 0.5*deg)
    n = 200
    detector = ba.SphericalDetector(n, -3*deg, 3*deg, n, 0, 4*deg)

    polarizer_vec = R3(0, 0, 1)
    analyzer_vec = R3(0, 0, -1)
    beam.setPolarization(polarizer_vec)
    detector.setAnalyzer(analyzer_vec)

    return ba.ScatteringSimulation(beam, sample, detector)


if __name__ == '__main__':
    sample = get_sample()
    ba.showSample3D(sample, sample_size=120*nm, seed=0)
    simulation = get_simulation(sample)
    result = simulation.simulate()
    bp.plot_datafield(result, unit_aspect=1)
    bp.plt.show()



#!/usr/bin/env python3
"""
Polarized neutron reflectometry of a simple magnetic layer.

Two-channel (PP, MM) angle-scan with Y-axis polarization.
Compare PolarizedQAngleReflectivity.py for the Q-space equivalent.
"""
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm, R3


def get_sample():
    vacuum = ba.Vacuum()
    layer_color = (0.86, 0.24, 0.18)
    layer_mat = ba.SLDMaterial("MagLayer", layer_color, 1e-4, 1e-8,
                                    R3(0, 1e8, 0))
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("Substrate", substrate_color, 7e-5, 2e-6)

    sample = ba.Sample()
    sample.addLayer(ba.Layer(vacuum))
    sample.addLayer(ba.Layer(layer_mat, 10*nm))
    sample.addLayer(ba.Layer(substrate_mat))
    return sample


def simulate(sample, pol, an, title):
    n = 500
    scan = ba.AlphaScan(n, 5*deg/n, 5*deg)
    scan.setWavelength(1.54*angstrom)
    scan.setPolarization(pol)
    scan.setAnalyzer(an)
    result = ba.SpecularSimulation(scan, sample).simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':
    sample = get_sample()
    results = [
        simulate(sample, R3(0, +1, 0), R3(0, +1, 0), "PP"),
        simulate(sample, R3(0, -1, 0), R3(0, -1, 0), "MM"),
    ]
    plotargs = bp.parse_commandline()
    bp.plot_multicurve(results, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
Cylinder form factor in DWBA with background/noise
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    from bornagain import std_samples
    return std_samples.cylinders()


def get_simulation(sample, background):
    beam = ba.Beam(1e8, 0.1*nm, 0.2*deg)
    n = 100
    detector = ba.SphericalDetector(n, 0., 2*deg, n, 0., 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    simulation.setBackground(background)
    return simulation


def simulate(background, title):
    sample = get_sample()
    simulation = get_simulation(sample, background)
    result = simulation.simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':    
    results = [
        simulate(ba.ConstantBackground(1e0), "Constant background"),
        simulate(ba.PoissonBackground(), "Poisson noise"),
    ]    
    plotargs = bp.parse_commandline()
    plotargs.update({'log_range': 5, 'unit_aspect': 1})
    bp.plot2d_to_row(results, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
Off-specular simulation showing a prominent Yoneda peak.

Sample: vacuum V | mixed M | substrate S.

Layer M contains a polydisperse 2D random assembly of cylindrical
particles P in vacuum.

Critical angle between vacuum and layer M: 2 deg.

The refractive index of substrate S equals the effective medium
index of layer M (linear mixing rule) so that we get no reflection
from the M/S interface. This leaves the V/M Yoneda peak as the
dominant feature.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm

# Target critical angle V->M: alpha_c = 2 deg
# delta_eff = alpha_c**2 / 2 = (2*pi/180)**2 / 2 = 6.09e-4
# Ambient = vacuum (delta_A=0); linear mixing at f=0.5:
#   delta_eff = f*delta_P  =>  delta_P = 2*delta_eff
DELTA_P   = 1.218e-3          # 2*DELTA_EFF
DELTA_EFF = 0.5*DELTA_P       # = 6.09e-4
H         = 3*nm               # layer thickness = cylinder height


def get_sample():
    particle_mat  = ba.RefractiveMaterial("Particle",  (0.86, 0.24, 0.18),
                                          DELTA_P, 0)
    substrate_mat = ba.RefractiveMaterial("Substrate", (0.28, 0.57, 0.82),
                                          DELTA_EFF, 0)

    # R = 5±1 nm cylinders, H=2 nm; polydispersity smears F(q_z) oscillations
    vacuum_layer    = ba.Layer(ba.Vacuum())
    mixed_layer     = ba.Layer(ba.Vacuum(), H)
    ff = ba.Cylinder(5*nm, H)
    particle = ba.Particle(particle_mat, ff)
    mixed_layer.deposit2D(ba.Dense2D(0.006, particle))
    substrate_layer = ba.Layer(substrate_mat)

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(mixed_layer)
    sample.addLayer(substrate_layer)
    return sample


def get_simulation(sample):
    nscan = <%= test_mode ? 50 : 200 %>
    scan = ba.AlphaScan(nscan, 1.6*deg, 3.4*deg)
    scan.setIntensity(1e8)
    scan.setWavelength(0.1*nm)
    detector = ba.OffspecDetector(nscan, 1.6*deg, 3.4*deg, -.2*deg, +.2*deg)
    return ba.OffspecSimulation(scan, sample, detector)


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    <%- if test_mode -%>
    from bornagain import ba_check
    ba_check.persistence_test(result)
    <%- elsif figure_mode -%>
    plotargs = bp.parse_commandline()
    bp.plot_datafield(result, intensity_min=1, **plotargs)
    bp.export(**plotargs)
    <%- else -%>
    ba.showSample3D(sample, sample_size=200*nm, seed=0)
    bp.plot_datafield(result)
    bp.plt.show()
    <%- end -%>



"""
BornAgain Python tools to inspect or modify sample.
"""
import bornagain as ba


def materialProfile(sample, n_points=400, z_min=None, z_max=None):
    """
    Creates a material profile from the given sample. If no limits are given,
    it will provide sensible default values, considering the included particles and
    interface roughnesses.
    :param sample: bornagain.Sample object
    :param n_points: number of points to generate
    :param z_min: starting value for z
    :param z_max: ending value for z
    :return: numpy arrays containing z positions and the complex material values in those positions
    """
    def_z_min, def_z_max = ba.defaultMaterialProfileLimits(sample)
    z_min = def_z_min if z_min is None else z_min
    z_max = def_z_max if z_max is None else z_max
    z_points = ba.generateZValues(n_points, z_min, z_max)
    values_mat = ba.materialProfileSLD(sample, n_points, z_min, z_max)
    return (z_points, values_mat)



"""
Check automatic slicing by comparing scattering from particle in decorated_layer
- (A) without slicing
- (B) with slicing provoked by dividing decorated_layer into sublayers
"""

import unittest
import PyFuTestInfrastructure as infrastruct
import bornagain as ba

matVacuum = ba.Vacuum()
matSubstrate_color = (0.55, 0.55, 0.55)
matSubstrate = ba.RefractiveMaterial("Substrate", matSubstrate_color, 3.2e-6, 3.2e-8)
matParticle_color = (0.86, 0.24, 0.18)
matParticle = ba.RefractiveMaterial("Ag", matParticle_color, 1.2e-5, 5.4e-7)

R = 10.0
dz = 4.0  # shift beneath interface


def get_sample(particle, sliced):
    """
    Returns a sample, with given particles attached to substrate or vacuum layer.
    """
    sample = ba.Sample()
    sample.addLayer(ba.Layer(matVacuum))
    if sliced:
        sample.addLayer(ba.Layer(matSubstrate, 5))
        sample.addLayer(ba.Layer(matSubstrate, 3))
        sample.addLayer(ba.Layer(matSubstrate, 2))
    else:
        sample.addLayer(ba.Layer(matSubstrate, 10))
    substrate = ba.Layer(matSubstrate)
    substrate.suspend2D(ba.Dilute2D(0.001, particle))
    sample.addLayer(substrate)

    return sample


class SlicedSpheresTest(unittest.TestCase):

    def runPlainFF(self, ff, eps = 1e-14):
        particle = ba.Particle(matParticle, ff)
        d1 = infrastruct.run_simulation_MiniGISAS(get_sample(particle, False))
        d2 = infrastruct.run_simulation_MiniGISAS(get_sample(particle, False))
        self.assertTrue(ba.checkRelativeDifference(d1, d2, eps))

    def testSlicingPlainFF(self):
        self.runPlainFF(ba.Cone(8., 9., 80*ba.deg))
        self.runPlainFF(ba.Cylinder(3., 9.))
        self.runPlainFF(ba.EllipsoidalCylinder(3., 4., 9.))
        self.runPlainFF(ba.Ellipsoid(7., 8., 9.))
        self.runPlainFF(ba.HorizontalCylinder(5., 19.), 5e-12)
        self.runPlainFF(ba.HorizontalCylinder(5., 19., 1., 1.), 5e-12)
        self.runPlainFF(ba.HorizontalCylinder(6., 19., 3., 3.), 5e-12)
        self.runPlainFF(ba.Sphere(5.))
        self.runPlainFF(ba.Spheroid(4., 4.))
        self.runPlainFF(ba.SphericalSegment(7., 9., 1.))
        self.runPlainFF(ba.SpheroidalSegment(7., 9., .9, 1.))
        self.runPlainFF(ba.EllipsoidalSegment(5., 7., 9., .9, 1.))

        self.runPlainFF(ba.PlatonicTetrahedron(12.)) # yet unsupported
        self.runPlainFF(ba.Prism3(8., 10.))
        self.runPlainFF(ba.Prism6(4., 9.))
        self.runPlainFF(ba.Pyramid3(8., 10., 80*ba.deg))
        self.runPlainFF(ba.Pyramid4(8., 9., 80*ba.deg))
        self.runPlainFF(ba.Pyramid6(18., 8., 40*ba.deg))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

"""
Various checks for certain C++ style rules that we choose for BornAgain,
in addition to the rules imposed by clang-format.

Copyright Forschungszentrum Jülich GmbH 2025.
License:  Public Domain
"""

import check_utils as cu
import os, re, sys

####################################################################################################
# checks
####################################################################################################

def rm_namespace(t):
    return re.sub(r'\nnamespace ([\w:]+ )?{\n(.*\n)*?}.*', r'\nnamespace /*...*/', t)

def check_block(fn, t, name, errs):

    if name == "include":
        rex = r'(#include) .*?'
    elif name == "using":
        rex = r'(using) [^;]*?;'
        t = rm_namespace(t)
    elif name == "fwd decl":
        rex = r'(class|enum class|struct) \w+;'
        t = rm_namespace(t)
    matches = [m for m in re.finditer(r'(\n*|template.*(\n\s*requires.*)*)\n('+ rex + r'([ ]*//.*)?\n)+(\n*)', t)]

    if len(matches) > 1:

        # tolerate multiple #include blocks if they are separated by some #if or #pragma
        if name == "include" and re.search('#include.*#(if|pragma).*#include', t, re.DOTALL):
            return

        errs.append(f"several {name} blocks in file {fn}")

        return

    elif len(matches) == 0:
        return

    m = matches[0]

    if re.match(r'^template', m.group(1)):
        return
    if m.group(1) == '':
        if not re.match(rex, t):
            errs.append(f"missing blank line above {name} block in file {fn}")
        else:
            print(f'In file {fn} block {name} at head')
    elif m.group(1) != '\n':
        errs.append(f"more than one blank line above {name} block in file {fn}")
    if m.group(6) == '':
        errs.append(f"missing blank line below {name} block in file {fn}")
    elif m.group(6) != '\n':
        errs.append(f"more than one blank line below {name} block in file {fn}")

def check_fwd_decl_sorted(fn, t, errs):
    for m in re.finditer(r'((class|struct) (\w+);([ ]*//.*)?\n)+', t):
        decls_txt = m.group(0)
        decls_arr = decls_txt.rstrip('\n').split('\n')
        if decls_arr != sorted(decls_arr):
            errs.append(f"forward declarations not sorted in file {fn}")
            return

def check_no_trivial_getters_in_cppfile(fc, t, errs):
    m = re.search(r'\n\S+ (\w+::[a-z]\w*)\(\) const\n{\s*return m_\w+;\s*}', tc)
    if m:
        errs.append(f"trivial getter {m.group(1)} in file {fc}, should go to header file")

####################################################################################################
# main
####################################################################################################

if __name__ == "__main__":

    assert len(sys.argv) == 1, "Check script must be called without arguments"

    errs = []

    # tests on each file:
    for fn in cu.all_files():
        stem, suffix = cu.split_fname(fn)
        with open(fn, 'r') as fd:
            t = fd.read()
        check_block(fn, t, 'include', errs)
        check_block(fn, t, 'using', errs)
        if suffix == 'h':
            check_block(fn, t, 'fwd decl', errs)
            check_fwd_decl_sorted(fn, t, errs)

    if len(errs) > 0:
        print("Found error(s):")
        for err in errs:
            print("- " + err)
        sys.exit(1)

    sys.exit(0)



#!/usr/bin/env python3
"""
Dilute soft spheres (FuzzySphere) in a thick film, 3D distribution.

Soft particles have diffuse boundaries modelled by a Debye-Waller factor.
Slicing is not supported for soft particles; material averaging is disabled.

The 3D distribution uses fill3D(Dilute3D).
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm3


def get_sample():
    R = 5*nm
    thickness = 12*R  # 60 nm

    # Materials
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-04, 2e-08)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Soft particle (FuzzySphere with Gaussian fuzziness), centered alignment
    ff = ba.FuzzySphere(R, 0.5*nm)
    particle = ba.Particle(particle_mat, ff)

    # Layers: vacuum / thick film / substrate
    layer_vacuum = ba.Layer(vacuum)
    layer_film = ba.Layer(vacuum, thickness)
    layer_film.fill3D(ba.Dilute3D(5e-4/nm3, particle))
    layer_substrate = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_vacuum)
    sample.addLayer(layer_film)
    sample.addLayer(layer_substrate)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 11
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 3*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    # Disable material averaging (slicing not supported for soft particles)
    simulation.options().setUseAvgMaterials(False)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#!/usr/bin/env python3
"""
This simulation example demonstrates how to replicate the
fitting example "Magnetically Dead Layers in Spinel Films"
given at the Nist website:
https://www.nist.gov/ncnr/magnetically-dead-layers-spinel-films

For simplicity, here we only reproduce the first part of that
demonstration without the magnetically dead layer.
"""

import os
import numpy
import bornagain as ba
from bornagain import nm, ba_plot as bp, deg, R3

# q-range on which the simulation and fitting are to be performed
qmin = 0.05997
qmax = 1.96

# number of points on which the computed result is plotted
scan_size = 1500

# The SLD of the substrate is kept constant
sldMao = (5.377e-06, 0)

# constant to convert between B and magnetic SLD
RhoMconst = 2.910429812376859e-12

####################################################################
#                  Create Sample and Simulation                    #
####################################################################

def get_sample(P):
    """
    construct the sample with the given parameters
    """
    BMagnitude = P["rhoM_Mafo"]*1e-6/RhoMconst
    angle = 0
    B = R3(
        BMagnitude*numpy.sin(angle*deg),
        BMagnitude*numpy.cos(angle*deg), 0)

    vacuum = ba.Vacuum()
    layer_color = (0.45, 0.32, 0.80)
    layer_mat = ba.SLDMaterial(
        "(Mg,Al,Fe)3O4", layer_color, P["rho_Mafo"]*1e-6, 0, B)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("MgAl2O4", substrate_color, *sldMao)

    r_Mafo_autocorr = ba.SelfAffineFractalModel(P["r_Mafo"]*nm, 0.7, 25*nm)
    r_substrate_autocorr = ba.SelfAffineFractalModel(P["r_Mao"]*nm, 0.7, 25*nm)

    transient = ba.TanhTransient()

    r_Mafo = ba.Roughness(r_Mafo_autocorr, transient)
    r_substrate = ba.Roughness(r_substrate_autocorr, transient)

    ambient_layer = ba.Layer(vacuum)
    layer = ba.Layer(layer_mat, P["t_Mafo"]*nm, r_Mafo)
    substrate_layer = ba.Layer(substrate_mat, r_substrate)

    sample = ba.Sample()
    sample.addLayer(ambient_layer)
    sample.addLayer(layer)
    sample.addLayer(substrate_layer)

    return sample


def get_simulation(sample, q_axis, parameters, polarizer_vec,
                   analyzer_vec):
    """
    A simulation object.
    Polarization, analyzer and resolution are set
    from given parameters
    """
    q_axis = q_axis + parameters["q_offset"]
    distr = ba.DistributionGaussian(0., 1., 25, 4.)

    scan = ba.QzScan(q_axis)
    scan.setAbsoluteQResolution(distr, parameters["q_res"])
                                       # TODO CHECK not parameters["q_res"]*q_axis ??

    scan.setPolarization(polarizer_vec)
    scan.setAnalyzer(analyzer_vec)

    return ba.SpecularSimulation(scan, sample)


def run_simulation(q_axis, fitP, *, polarizer_vec, analyzer_vec):
    """
    Run a simulation on the given q-axis, where the sample is
    constructed with the given parameters.
    Vectors for polarization and analyzer need to be provided
    """
    parameters = dict(fitP, **fixedP)

    sample = get_sample(parameters)
    simulation = get_simulation(sample, q_axis, parameters, polarizer_vec,
                                analyzer_vec)

    return simulation.simulate()


def qr(result):
    """
    Returns two arrays that hold the q-values as well as the
    reflectivity from a given simulation result
    """
    q = result.xCenters()
    r = result.intensities()

    return q, r

####################################################################
#                         Plot Handling                            #
####################################################################

def plotData(qs, rs, exps, labels, colors):
    """
    Plot the simulated result together with the experimental data
    """
    fig = bp.plt.figure()
    ax = fig.add_subplot(111)

    for q, r, exp, l, c in zip(qs, rs, exps, labels, colors):
        ax.errorbar(exp.xCenters(),
                    exp.intensities(),
                    # xerr=TODO i742,
                    yerr=exp.errorSigmas(),
                    fmt='.',
                    markersize=0.75,
                    linewidth=0.5,
                    color=c[1])
        ax.plot(q, r, label=l, color=c[0])

    ax.set_yscale('log')
    bp.plt.legend()

    bp.plt.xlabel(r"$q$ (nm$^{-1}$)")
    bp.plt.ylabel("$R$")

    bp.plt.tight_layout()

def plotSpinAsymmetry(data_pp, data_mm, q, r_pp, r_mm):
    """
    Plot the simulated spin asymmetry as well its
    experimental counterpart with errorbars
    """
    Yp = data_pp.intensities()
    Ym = data_mm.intensities()
    Ep = data_pp.errors()
    Em = data_mm.errors()
    # compute the errorbars of the spin asymmetry
    delta = numpy.sqrt(4 * (Yp**2 * Em**2 + Ym**2 * Ep**2 ) / ( Yp + Ym )**4 )

    fig = bp.plt.figure()
    ax = fig.add_subplot(111)

    ax.errorbar(data_pp.xCenters(),
                (Yp - Ym) / (Yp + Ym),
                # xerr=TODO i742,
                yerr=delta,
                fmt='.',
                markersize=0.75,
                linewidth=0.5)

    ax.plot(q, (r_pp - r_mm)/(r_pp + r_mm))

    bp.plt.gca().set_ylim((-0.3, 0.5))

    bp.plt.xlabel(r"$q$ (nm$^{-1}$)")
    bp.plt.ylabel("Spin asymmetry")

    bp.plt.tight_layout()

####################################################################
#                          Data Handling                           #
####################################################################

def load_data(fname):
    flags = ba.ImportSettings1D("q_z (1/nm)", "", "", 1, 2, 3, 4)
    return ba.readData1D(fname, ba.csv1D, flags)

####################################################################
#                          Main Function                           #
####################################################################

if __name__ == '__main__':
    datadir = os.getenv('BA_DATA_DIR', '')
    if not datadir:
        raise Exception("Environment variable BA_DATA_DIR not set")
    fname_stem = os.path.join(datadir, "specular/MAFO_Saturated_")

    expdata_pp = load_data(fname_stem + "pp.tab")
    expdata_mm = load_data(fname_stem + "mm.tab")

    fixedP = {
        # parameters from our own fit run
        'q_res': 0.010542945012551425,
        'q_offset': 7.971243487467318e-05,
        'rho_Mafo': 6.370140108715461,
        'rhoM_Mafo': 0.27399566816062926,
        't_Mafo': 13.746913056084736,
        'r_Mao': 0.860487712674644,
        'r_Mafo': 0.37844265311293483
    }

    def run_Simulation_pp(qzs, P):
        return run_simulation(qzs,
                              P,
                              polarizer_vec=R3(0, 1, 0),
                              analyzer_vec=R3(0, 1, 0))

    def run_Simulation_mm(qzs, P):
        return run_simulation(qzs,
                              P,
                              polarizer_vec=R3(0, -1, 0),
                              analyzer_vec=R3(0, -1, 0))

    qzs = numpy.linspace(qmin, qmax, scan_size)
    ba.showSample3D(get_sample(fixedP), sample_size=120*nm, seed=0)
    q_pp, r_pp = qr(run_Simulation_pp(qzs, fixedP))
    q_mm, r_mm = qr(run_Simulation_mm(qzs, fixedP))

    plotData([q_pp, q_mm], [r_pp, r_mm], [expdata_pp, expdata_mm],
             ["$++$", "$--$"], [['orange','red'], ['green','blue']])

    plotSpinAsymmetry(expdata_pp, expdata_mm, qzs, r_pp, r_mm)
    bp.plt.show()



#!/usr/bin/env python3
"""
GISAS of sawtooth ripples on a 2D paracrystal lattice.

SawtoothRippleBox (100 x 20 x 10 nm, asymmetry d=-3 nm) rotated 30 deg
around Z, arranged on a BasicLattice2D (20 x 100 nm, 90 deg angle,
30 deg rotation) using Paracrystal2D with Gaussian pair-distance
distributions (sigma=150 x 600 nm, orientation 0 deg).
"""
import math
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-4, 2e-8)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-6, 2e-8)
    vacuum = ba.Vacuum()

    length_ripple = 100*nm
    ff = ba.SawtoothRippleBox(length_ripple, 20*nm, 10*nm, -3*nm)
    particle = ba.Particle(particle_mat, ff)
    particle.rotate(ba.RotationZ(math.pi/6))

    lattice = ba.BasicLattice2D(20*nm, length_ripple, math.pi/2, math.pi/6)
    structure = ba.Paracrystal2D(particle, lattice, 1000*nm, 1e4*nm, 1e4*nm)
    profile = ba.Profile2DGauss(150*nm, 600*nm, 0)
    structure.setProbabilityDistributions(profile, profile)

    vacuum_layer = ba.Layer(vacuum)
    substrate_layer = ba.Layer(substrate_mat)
    vacuum_layer.deposit2D(structure)

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(substrate_layer)
    return sample


def get_simulation(sample):
    beam = ba.Beam(1, 0.1*nm, 0.2*deg)
    n = <%= test_mode ? 25 : 100 %>
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 2*deg)
    return ba.ScatteringSimulation(beam, sample, detector)


<%= include_partial("template/main/runsim_unitaspect.py",
    "ba.showSample3D(sample, sample_size=1000*nm, seed=0)") -%>



#!/usr/bin/env python3
"""
GISAS of a centered square lattice (C2mm) with a two-cylinder basis.

The unit cell contains two cylinders at (0,0,0) and (5,-5,0) nm on a
20 nm × 20 nm square lattice, forming a centered (C2mm) arrangement.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3
import math


def get_sample():
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-4, 2e-8)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-6, 2e-8)
    vacuum = ba.Vacuum()

    ff = ba.Cylinder(5 * nm, 5 * nm)
    cylinder = ba.Particle(particle_mat, ff)

    basis = ba.Compound()
    basis.addComponents(cylinder, [R3(0, 0, 0), R3(5 * nm, -5 * nm, 0)])

    lattice = ba.BasicLattice2D(20 * nm, 20 * nm, 90 * deg, 0)
    structure = ba.Crystal2D(basis, lattice)
    profile = ba.Profile2DCauchy(300 / (2 * math.pi) * nm,
                                 100 / (2 * math.pi) * nm, 0)
    structure.setDecayFunction(profile)

    vacuum_layer = ba.Layer(vacuum)
    substrate_layer = ba.Layer(substrate_mat)
    vacuum_layer.deposit2D(structure)

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(substrate_layer)
    return sample


<%= include_partial("template/sim/gisas_standard_wide.py") -%>


<%= include_partial("template/main/runsim_unitaspect.py",
    "ba.showSample3D(sample, sample_size=240*nm, seed=0)") -%>



#!/usr/bin/env python3
"""
GISAS of two boxes rotated first around Z then around Y.

Two half-length boxes (25 × 20 × 10 nm) form a composition, rotated
sequentially 90° around Z then 90° around Y, and embedded in a 100 nm
Teflon layer. Material averaging is disabled.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3

layer_thickness = 100 * nm
length = 50 * nm
width = 20 * nm
height = 10 * nm


def get_sample():
    vacuum = ba.Vacuum()
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Ag", particle_color, 1.245e-5, 5.419e-7)
    teflon_color = (0.93, 0.72, 0.25)
    teflon_mat = ba.RefractiveMaterial("Teflon", teflon_color, 2.9e-6, 6.019e-9)
    substrate_color = (0.24, 0.50, 0.74)
    substrate_mat = ba.RefractiveMaterial(
        "Substrate2", substrate_color, 3.212e-6, 3.244e-8)

    box = ba.Particle(particle_mat, ba.Box(length / 2, width, height))
    composition = ba.Compound()
    composition.addComponent(box, R3())
    composition.addComponent(box, R3(length / 2, 0, 0))
    composition.rotate(ba.RotationZ(90 * deg))
    composition.rotate(ba.RotationY(90 * deg))
    composition.translate(R3(0, 0, -layer_thickness / 2))

    vacuum_layer = ba.Layer(vacuum)
    middle_layer = ba.Layer(teflon_mat, layer_thickness)
    substrate_layer = ba.Layer(substrate_mat)
    vacuum_layer.deposit2D(ba.Dilute2D(5e-4, composition))

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(middle_layer)
    sample.addLayer(substrate_layer)
    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1 * nm, 0.2 * deg)
    n = 200
    detector = ba.SphericalDetector(n, -2 * deg, 2 * deg, n, 0, 2 * deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    simulation.options().setUseAvgMaterials(False)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    plotargs = bp.parse_commandline()
    bp.plot_datafield(result, unit_aspect=1, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
GISAS from a periodic multilayer with self-affine fractal roughness.

Five repetitions of [PartA (2.5 nm) / PartB (5 nm)] on a substrate,
with correlated roughness described by a self-affine fractal model and
tanh transition profile.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    vacuum = ba.Vacuum()
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 15e-6, 0)
    part_a_color = (0.86, 0.24, 0.18)
    part_a_mat = ba.RefractiveMaterial("PartA", part_a_color, 5e-6, 0)
    part_b_color = (0.25, 0.65, 0.35)
    part_b_mat = ba.RefractiveMaterial("PartB", part_b_color, 10e-6, 0)

    autocorr = ba.SelfAffineFractalModel(1 * nm, 0.3, 5 * nm)
    transient = ba.TanhTransient()
    crosscorr = ba.CommonDepthCrosscorrelation(10 * nm)
    roughness = ba.Roughness(autocorr, transient, crosscorr)

    vacuum_layer = ba.Layer(vacuum)
    partA_layer = ba.Layer(part_a_mat, 2.5 * nm, roughness)
    partB_layer = ba.Layer(part_b_mat, 5 * nm, roughness)
    substrate_layer = ba.Layer(substrate_mat, roughness)

    stack = ba.LayerStack(5)
    stack.addLayer(partA_layer)
    stack.addLayer(partB_layer)

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addStack(stack)
    sample.addLayer(substrate_layer)
    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 11
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#!/usr/bin/env python3
"""
Specular reflectometry with hand-averaged SLD layers (no particles).

Instead of explicit particles with setNumberOfSlices, N homogeneous SLD
layers with the volume-averaged scattering length density replace the
sliced-particle layer. Results must agree with SpecularWithSlicing2.py.
"""
from math import pi
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm

HEIGHT = 5 * nm
RADIUS = 5 * nm
SURF_DENSITY = 0.01   # nm^-2
N_SLICES = 3
WAVELENGTH = 1.54 * angstrom


def _sld(delta, beta):
    """Convert refractive index (delta, beta) to SLD (Å⁻²) at WAVELENGTH."""
    f = pi * angstrom**2 / WAVELENGTH**2
    return (2*delta - delta**2 + beta**2) * f, (2*beta - 2*delta*beta) * f


def get_sample():
    vacuum = ba.Vacuum()
    sub_re, sub_im = _sld(6e-6, 2e-8)
    par_re, par_im = _sld(6e-4, 2e-8)

    # Effective volume fraction of cylinders in each slice
    eff_vol = SURF_DENSITY * pi * RADIUS**2
    avr_re = eff_vol * par_re
    avr_im = eff_vol * par_im

    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("Substrate", substrate_color, sub_re, sub_im)
    avr_color = (0.05, 0.62, 0.55)
    avr_mat = ba.SLDMaterial("Avr", avr_color, avr_re, avr_im)

    sample = ba.Sample()
    sample.addLayer(ba.Layer(vacuum))
    for _ in range(N_SLICES):
        sample.addLayer(ba.Layer(avr_mat, HEIGHT / N_SLICES))
    sample.addLayer(ba.Layer(substrate_mat))
    return sample


def get_simulation(sample):
    n = 500
    scan = ba.AlphaScan(n, 5*deg/n, 5*deg)
    scan.setWavelength(WAVELENGTH)
    return ba.SpecularSimulation(scan, sample)


if __name__ == '__main__':
    sample = get_sample()
    ba.showSample3D(sample, sample_size=120*nm, seed=0)
    simulation = get_simulation(sample)
    result = simulation.simulate()
    bp.plot_datafield(result)
    bp.plt.show()



#!/usr/bin/env python3
"""
Check internal links in md file in or under hugo directory.
Used by BornAgain CI to ensure that all internal links work in the online documentation.

Copyright Forschungszentrum Julich GmbH 2025.
License:  Public Domain
"""
import glob, os, re, sys

# Shortcode file path mappings
# Maps (shortcode_name, parameter_name) to path transformation function
# Paths starting with / are treated as absolute from hugo/content
SHORTCODE_MAPPINGS = {
    ('figscg', 'src'): lambda p: p,  # Direct path, already starts with /
    ('show-ex', 'file'): lambda p: '/static/py/auto/Examples/' + p,
    ('ref-ex', None): lambda p: '/auto/Examples/' + p,  # Absolute from repo root
    ('ref-data-scat2d', None): lambda p: '/testdata/scatter2d/' + p.strip(),
    ('ref-data-spec', None): lambda p: '/testdata/specular/' + p.strip(),
    ('ref-src', None): lambda p: '/' + p,  # Absolute from repo root
    ('ref-class', None): lambda parts: '/' + parts[0] + '/' + parts[1] + '.h',  # Two args
}


def print_broken_links(first_line_printed, broken_links_count, lin,
                       links_list, src_list):
    '''
    Increase the broken links counter by 1 and return it in the output.
    Print the broken links with the corresponding md files in relative form.
    Change first_line_printed in case it was False to indicate that the header
    is printed and return its value.
    '''
    if not first_line_printed:
        print("Broken link: ")
        first_line_printed = True
    broken_links_count += 1
    print('\n', lin)
    print("In: ")
    for i in range(len(links_list)):
        if links_list[i] == lin:
            print(src_list[i])
    return (first_line_printed, broken_links_count)


def make_path_absolute(current_path, relative_path):
    '''
    This is to transform relative path into an absolute one
    example: current_path is ./ref/sim/class, relative_path: ../../file
    as long as there is ../ in the relative path, it delete them, but
    at the same time it delete one level from the current path
    inside while, in the first iteration: relative path becomes ../file
    and current_path becomes ./ref/sim
    '''
    while relative_path[0:3] == '../':
        relative_path = relative_path[3:len(relative_path)]
        current_path = current_path[0:int(current_path.rfind('/'))]
    return f'{current_path}/{relative_path}'


def check_one_file(mdf):
    '''
    Read a file (mdf).
    Check the searched pattern in a file and send each to test.
    '''
    with open(mdf, 'r', encoding="utf8") as f:
        try:
            ftxt = f.read()
        except Exception as e:
            print('A problem appeard while trying to read the file: ' +
                  mdf)
            print(e)
            exit(1)
        # Check regular markdown links
        for caught in re.findall(r'\[.*?\]\((.*?)\)', ftxt):
            check_one_link(mdf, caught)

        # Check shortcode links using the mapping
        # Named parameters with {{< >}} syntax
        for caught in re.findall(r'{{<\s*figscg\s+src="([^"]+)"', ftxt):
            check_one_link(mdf, SHORTCODE_MAPPINGS[('figscg', 'src')](caught))
        for caught in re.findall(r'{{<\s*show-ex\s+file="([^"]+)"', ftxt):
            check_one_link(mdf, SHORTCODE_MAPPINGS[('show-ex', 'file')](caught))

        # Positional parameters with {{% %}} syntax
        for caught in re.findall(r'{{% ref-ex "([^"]+)" %}}', ftxt):
            check_one_link(mdf, SHORTCODE_MAPPINGS[('ref-ex', None)](caught))
        for caught in re.findall(r'{{% ref-data-scat2d ([^ %}]+) %}}', ftxt):
            check_one_link(mdf, SHORTCODE_MAPPINGS[('ref-data-scat2d', None)](caught))
        for caught in re.findall(r'{{% ref-data-spec ([^ %}]+) %}}', ftxt):
            check_one_link(mdf, SHORTCODE_MAPPINGS[('ref-data-spec', None)](caught))
        for caught in re.findall(r'{{% ref-src "([^"]+)" %}}', ftxt):
            check_one_link(mdf, SHORTCODE_MAPPINGS[('ref-src', None)](caught))
        for caught in re.findall(r'{{% ref-class "([^"]+)" "([^"]+)" %}}', ftxt):
            check_one_link(mdf, SHORTCODE_MAPPINGS[('ref-class', None)](caught))

def check_one_link(mdf, caught):
    '''
    Check a one detected pattern (caught) in a file (mdf)
    Checking the format of caught, if it is a string, if it contains special characters,..
    Transform conformal link to a unified form
    Append mdf and link to src_list and to Links_list.
    '''
    try:
        line = str(caught)
    except Exeption as e:
        print('The following error appeard when dealing with file: ')
        print(mdf)
        print('It happened when reading the link as a string')
        print(e)
        exit(1)
    # Strip title attribute if present (e.g., "path/file.png "title text"")
    if ' "' in line:
        line = line[:line.index(' "')]
    # Skip empty strings
    if not line:
        return
    if '.md' in line:
        dotmdlst.append(line)
    if line[0:1] == '/':
        tline = '.' + line
        src_list.append(mdf)
        links_list.append(tline)
    elif line[0:4] == 'http':
        pass  # don't check external links for the time being
    elif line[0] == '#':
        pass  # link target is inside present page - no need for checking
    elif line[0:2] == '{{':
        pass  # TODO, shortcodes to be treated later
    elif line[0:7] == 'mailto:':
        pass
    else:
        pp = mdf[0:int(mdf.rfind('/'))]
        tline = make_path_absolute(pp, line)
        src_list.append(mdf)
        links_list.append(tline)


def check_existence(lin, broken_links_count):
    '''
    Check if the link is broken, print related information if it is the case
    increase broken links count when neccessay
    '''
    first_line_printed = False
    if '#' in lin:
        lin = lin[0:int(lin.rfind('#'))]
    if '.md' in lin:
        dotmdlst.append(lin)
    # Check if this is a static file link (starts with ./img/, ./css/, ./static/, etc.)
    # Static files are in hugo/static/, not hugo/content/
    # Repository files (auto/, testdata/, Sample/, etc.) are in repo root ../
    check_path = lin
    if lin.startswith('./img/') or lin.startswith('./css/') or lin.startswith('./js/') or lin.startswith('./files/'):
        # Convert ./img/... to ../static/img/...
        check_path = '../static/' + lin[2:]
    elif lin.startswith('./static/'):
        # Convert ./static/... to ../static/...
        check_path = '..' + lin[1:]
    elif (lin.startswith('./auto/') or lin.startswith('./testdata/') or lin.startswith('./Sample/') or
          lin.startswith('./scripts/') or lin.startswith('./devtools/') or lin.startswith('./Device/') or
          lin.startswith('./3rdparty/') or lin == './testdata' or lin == './3rdparty'):
        # Convert ./auto/... to ../../auto/... (repository root files, two levels up from hugo/content)
        check_path = '../..' + lin[1:]
    # Check if path exists as file or directory
    if os.path.isdir(check_path):
        # Path is a valid directory, no need to check further
        return broken_links_count
    if not os.path.exists(check_path):
        '''
        os.path.exists() does not give a correct answer when checking for a file
        without its extension.
        example when checking ./ref/example where example is an md file, it returns False
        The goal of the following is first separate the path from the name of the file
        pthchk (path check) in the example is to check ./ref
        fdirchk is the name example.
        Then we have to list the name of what pthchk contains without extensions
        if fdirchk match one of them then the link is not broken
        '''
        pthchk = check_path[0:int(check_path.rfind('/'))]
        fdirchk = check_path[int(check_path.rfind('/')) + 1:len(check_path)]
        direx = os.path.exists(pthchk)
        if not direx:
            (first_line_printed, broken_links_count) = print_broken_links(
                first_line_printed, broken_links_count, lin, links_list,
                src_list)
        elif direx:
            lsta = glob.glob(pthchk + '/*') # list pthchk directory items
            newl = [] # a list to append the same items but without extensions
            for elea in lsta:
                elea = elea[int(elea.rfind('/')) + 1:len(elea)] # here just to get the names without any path part
                if '.' in elea: # if it contains extension
                    elea = elea[0:int(elea.rfind('.'))]
                newl.append(elea.lower())
            if fdirchk not in newl:
                (first_line_printed,
                 broken_links_count) = print_broken_links(
                     first_line_printed, broken_links_count, lin,
                     links_list, src_list)
    return broken_links_count

src_list = []
links_list = []
dotmdlst = []

def check_hugo_links():
    '''
    The main function. it gather first all links in the absolute form,
    then it checks them
    '''
    mdlist = []    
    mdlist = glob.glob('./**/*.md', recursive=True)
    for ele in range(len(mdlist)):
        mdf = str(mdlist[ele])
        check_one_file(mdf)
    if links_list == []:
        print("Links list is empty")
        exit(1)
    disabs = list(set(links_list))
    total_links_count = len(disabs)
    broken_links_count = 0
    for cc in range(len(disabs)):
        link = disabs[cc]
        broken_links_count = check_existence(link, broken_links_count)
    print('--------------------------------------------------')
    print('links containing .md extension are: ', len(dotmdlst))
    print(
        f'{broken_links_count}/{total_links_count} links are broken in total!'
    )
    if broken_links_count > 0:
        exit(1)


if __name__ == '__main__':
    check_hugo_links()



#!/usr/bin/env python3
"""
GISAS of a 3D array of spheres enclosed in a cylindrical envelope.

Spheres of radius 4 nm are placed on a 10 nm cubic grid within a
cylinder of radius (n_lim+0.5)·L and height (2·n_lim+1)·L. The
resulting clusters are deposited at low density (Dilute2D).
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3


def get_sample():
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-4, 2e-8)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-6, 2e-8)
    vacuum = ba.Vacuum()

    n_lim = 1
    L = 10 * nm
    R = 4 * nm

    ff = ba.Sphere(R)
    sphere = ba.Particle(particle_mat, ff)

    size = L * (2 * n_lim + 1 - 1) + 2 * R
    outer_r = size / 2

    positions = []
    for i in range(-n_lim, n_lim + 1):
        for j in range(-n_lim, n_lim + 1):
            for k in range(0, 2 * n_lim + 1):
                x, y, z = i * L, j * L, k * L
                if x**2 + y**2 <= outer_r**2 and 0 <= z <= size:
                    positions.append(R3(x, y, z))

    basis = ba.Compound()
    basis.addComponents(sphere, positions)

    vacuum_layer = ba.Layer(vacuum)
    substrate_layer = ba.Layer(substrate_mat)
    vacuum_layer.deposit2D(ba.Dilute2D(0.001, basis))

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(substrate_layer)
    return sample


<%= include_partial("template/sim/gisas_standard_wide.py") -%>


<%= include_partial("template/main/runsim_unitaspect.py",
    "ba.showSample3D(sample, sample_size=250*nm, seed=0)") -%>



"""
BornAgain collection of standard simulation setups.
"""
import bornagain as ba
from bornagain import deg, angstrom


def particle(particle, n):
    """
    Returns a ParticleSimulation for computing pure form factor scattering.
    """
    beam = ba.Beam(1, 1*angstrom, 0)
    det = ba.SphericalDetector(n, -4.5*deg, 4.5*deg, n, -4.5*deg, 4.5*deg)
    return ba.ParticleSimulation(beam, particle, det)



#!/usr/bin/env python3
"""
Specular reflectometry with sliced cylinders (Q-scan variant).

Same sample as SpecularWithSlicing2.py but the scan axis is q_z
instead of incident angle.
"""
from math import pi, sin
import numpy as np
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm

HEIGHT = 5 * nm
RADIUS = 5 * nm
SURF_DENSITY = 0.01   # nm^-2
N_SLICES = 3
WAVELENGTH = 1.54 * angstrom


def _sld(delta, beta):
    """Convert refractive index (delta, beta) to SLD (Å⁻²) at WAVELENGTH."""
    f = pi * angstrom**2 / WAVELENGTH**2
    return (2*delta - delta**2 + beta**2) * f, (2*beta - 2*delta*beta) * f


def get_sample():
    vacuum = ba.Vacuum()
    sub_re, sub_im = _sld(6e-6, 2e-8)
    par_re, par_im = _sld(6e-4, 2e-8)

    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("Substrate", substrate_color, sub_re, sub_im)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.SLDMaterial("Particle", particle_color, par_re, par_im)

    cylinder = ba.Cylinder(RADIUS, HEIGHT)
    particle = ba.Particle(particle_mat, cylinder)

    top_layer = ba.Layer(vacuum)
    top_layer.deposit2D(ba.Dilute2D(SURF_DENSITY, particle))
    top_layer.setNumberOfSlices(N_SLICES)

    sample = ba.Sample()
    sample.addLayer(top_layer)
    sample.addLayer(ba.Layer(substrate_mat))
    return sample


def get_simulation(sample):
    n = 50
    lam = WAVELENGTH
    qs = [4*pi*sin(a)/lam for a in np.linspace(5*deg/n, 5*deg, n)]
    scan = ba.QzScan(qs)
    return ba.SpecularSimulation(scan, sample)


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#!/usr/bin/env python3
"""
Cylinder form factor in Born approximation, beam incident from below.

Cylinders (5 × 5 nm) in a vacuum matrix, with the beam angle set to
-0.2° (negative grazing angle), corresponding to a beam that enters
the sample from below the substrate. The detector is placed below the
horizon (alpha_f from -2° to 0°).
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-4, 2e-8)
    vacuum = ba.Vacuum()

    ff = ba.Cylinder(5*nm, 5*nm)
    particle = ba.Particle(particle_mat, ff)

    layer_top = ba.Layer(vacuum)
    layer_bottom = ba.Layer(vacuum)
    layer_top.deposit2D(ba.Dilute2D(0.001, particle))

    sample = ba.Sample()
    sample.addLayer(layer_top)
    sample.addLayer(layer_bottom)
    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, -0.2*deg)
    n = 200
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, -2*deg, 0)
    return ba.ScatteringSimulation(beam, sample, detector)


if __name__ == '__main__':
    sample = get_sample()
    ba.showSample3D(sample, sample_size=120*nm, seed=0)
    simulation = get_simulation(sample)
    result = simulation.simulate()
    bp.plot_datafield(result, unit_aspect=1)
    bp.plt.show()



#!/usr/bin/env python3
"""
Basic GISAS 2D fitting example.
The model is in gisas_model1.
Fake experimental data are generated by gisas_fake1.
"""

import os
import numpy as np
import gisas_model1 as model
import bornagain as ba
from bornagain import ba_fitmonitor, nm
import lmfit

from matplotlib import pyplot as plt


if __name__ == '__main__':
    datadir = os.getenv('BA_DATA_DIR', '')
    if not datadir:
        raise Exception("Environment variable BA_DATA_DIR not set")
    fname = os.path.join(datadir, "scatter2d/faked_gisas1.int")
    data = ba.readData2D(fname)

    fit_objective = ba.FitObjective()
    fit_objective.addFitPair(model.get_simulation, data)

    fit_objective.initPrint(10)  # Print on every 10th iteration.
    observer = ba_fitmonitor.PlotterGISAS()
    <%- if test_mode or figure_mode -%>
    plt.close() # hide plot
    <%- else -%>
    fit_objective.initPlot(10, observer) # Plot every 10th, slow!
    <%- end -%>

    P = model.start_parameters_1()
    result = lmfit.minimize(fit_objective.evaluate_residuals, P)

    print("Fit completed.")
    print(lmfit.fit_report(result))
    <%- unless test_mode || figure_mode -%>
    ba.showSample3D(model.get_sample(result.params), sample_size=120*nm, seed=0)
    <%- end -%>

    # Save simulation image corresponding to the best fit parameters
    np.savetxt("fit.txt",
               fit_objective.simulationResult().intensities())
    plt.show()



#!/usr/bin/env python3
"""
"""
import numpy as np
import bornagain as ba
from bornagain import ba_plot as bp, nm


def get_sample():
    # Materials
    a_color = (0.05, 0.62, 0.55)
    a_mat = ba.SLDMaterial("A", a_color, 5e-06, 0)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("Substrate", substrate_color, 2e-06, 0)

    # Layers
    ambient_layer = ba.Layer(ba.Vacuum())
    film = ba.Layer(a_mat, 30*nm)
    substrate_layer = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(ambient_layer)
    sample.addLayer(film)
    sample.addLayer(substrate_layer)

    return sample

def get_simulation(sample):
    "Specular simulation with a qz-defined beam"
    n = 50
    qzs = np.linspace(0.01, 1, n)  # qz-values
    scan = ba.QzScan(qzs)
    return ba.SpecularSimulation(scan, sample)


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#!/usr/bin/env python3
"""
Minimal working fit examples: finds radius of sphere in Born approximation.
"""
import bornagain as ba
from bornagain import deg, nm, nm2
import lmfit


def get_sample(P):
    """
    Returns sample for given set of parameters.
    """
    radius = P["radius"]

    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-4, 2e-8)
    particle = ba.Particle(particle_mat, ba.Sphere(radius))

    layer_top = ba.Layer(ba.Vacuum())
    layer_bottom = ba.Layer(ba.Vacuum())
    layer_top.deposit2D(ba.Dilute2D(0.01/nm2, particle))

    sample = ba.Sample()
    sample.addLayer(layer_top)
    sample.addLayer(layer_bottom)

    return sample


def get_simulation(P):
    """
    Returns GISAS simulation for given set of parameters.
    """
    sample = get_sample(P)

    n = 100
    beam = ba.Beam(1, 0.1*nm, 0.2*deg)
    detector = ba.SphericalDetector(n, -1*deg, 1*deg, n, 0., 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)

    return simulation


def fake_data():
    """
    Generating "experimental" data by running simulation with default parameters.
    """
    simulation = get_simulation({'radius': 5 * nm})
    result = simulation.simulate()
    return result


if __name__ == '__main__':
    data = fake_data()
    fit_objective = ba.FitObjective()
    fit_objective.addFitPair(get_simulation, data)
    fit_objective.initPrint(10)

    P = lmfit.Parameters()
    P.add("radius", 4. * nm, min=0.01)

    result = lmfit.minimize(fit_objective.evaluate_residuals, P)
    print(lmfit.fit_report(result))
    ba.showSample3D(get_sample(result.params), sample_size=120*nm, seed=0)



#!/usr/bin/env python3
"""
Flattened depth-probe simulation:
Intensity as function of depth z for a few incident angles alpha_i.
"""

import bornagain as ba
from bornagain import ba_plot as bp, deg, nm


def get_sample():
    vac_mat = ba.Vacuum()
    a_color = (0.05, 0.62, 0.55)
    a_mat = ba.RefractiveMaterial("A", a_color, 6e-5, 0)
    sub_color = (0.28, 0.57, 0.82)
    sub_mat = ba.RefractiveMaterial("Substrate", sub_color, 3e-05, 0)

    sample = ba.Sample()
    sample.addLayer(ba.Layer(vac_mat))
    sample.addLayer(ba.Layer(a_mat, 100*nm))
    sample.addLayer(ba.Layer(sub_mat))

    return sample

def simulate(sample, alpha_i_deg):
    alpha = alpha_i_deg * deg
    scan = ba.AlphaScan(1, alpha, alpha)
    scan.setWavelength(0.3*nm)

    z_axis = ba.EquiDivision("z (nm)", <%= test_mode ? 20 : 500 %>, -130*nm, 30*nm)
    simulation = ba.DepthprobeSimulation(scan, sample, z_axis, 0)

    result = simulation.simulate().flat()
    result.setTitle(f'{alpha_i_deg} deg')
    return result

if __name__ == '__main__':
    sample = get_sample()
    angles = [0.4, 0.5, 0.6, 0.65, 0.7]
    results = [simulate(sample, ai) for ai in angles]
    <%- if test_mode -%>
    from bornagain import ba_check
    ba_check.persistence_test(result)
    <%- elsif figure_mode -%>
    plotargs = bp.parse_commandline()
    bp.plot_multicurve(results, legendloc='lower right', **plotargs)
    bp.export(**plotargs)
    <%- else -%>
    ba.showSample3D(sample, sample_size=180*nm, seed=0)
    bp.plot_multicurve(results, legendloc='lower right')
    bp.plt.show()
    <%- end -%>



#!/usr/bin/env python3

"""
Utilities for code checker scripts.

Copyright Forschungszentrum Jülich GmbH 2025.
License:  Public Domain
"""

import re, subprocess, sys

def split_fname(fn):
    """
    return stem, suffix
    """
    m = re.match(r'(.*)\.(\w+)$', fn)
    if m:
        return (m.group(1), m.group(2))
    return (fn, "")

import os, re, sys

def all_files(root='.'):
    excluded_dirs = {
        './auto', './devtools', './build', './debug',
        './cover', './tidy', './qbuild'
    }
    excluded_patterns = re.compile(r'.*(ThirdParty|3rdparty).*')
    included_extensions = {'.cpp', '.c', '.h'}

    result_files = []
    for dirpath, dirnames, filenames in os.walk(root):
        # Normalize path
        dirpath = os.path.normpath(dirpath)

        # Skip excluded directories
        if any(os.path.commonpath([dirpath, os.path.normpath(ex)]) == os.path.normpath(ex) for ex in excluded_dirs):
            dirnames[:] = []  # Don't descend into this directory
            continue

        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            if excluded_patterns.search(filepath):
                continue
            if os.path.splitext(filename)[1] in included_extensions:
                result_files.append(filepath)

    return result_files



#!/usr/bin/env python3
"""
Specular reflectometry of a Ti/Ni multilayer with a non-unit beam
intensity factor. Demonstrates that angle-scan and Q-scan yield
consistent results when the same intensity factor is applied.

10 bilayers of Ti (3 nm) and Ni (7 nm) on a Si substrate.
"""
import numpy as np
from math import pi, sin
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm


INTENSITY = 1.05


def get_sample():
    vacuum = ba.Vacuum()
    si_color = (0.28, 0.57, 0.82)
    si_mat = ba.SLDMaterial("Si_substrate", si_color, 2.0704e-6, 2.3726e-11)
    ni_color = (0.93, 0.48, 0.14)
    ni_mat = ba.SLDMaterial("Ni", ni_color, 9.4245e-6, 1.1423e-9)
    ti_color = (0.05, 0.62, 0.55)
    ti_mat = ba.SLDMaterial("Ti", ti_color, -1.9493e-6, 9.6013e-10)

    stack = ba.LayerStack(10)
    stack.addLayer(ba.Layer(ti_mat, 3*nm))
    stack.addLayer(ba.Layer(ni_mat, 7*nm))

    sample = ba.Sample()
    sample.addLayer(ba.Layer(vacuum))
    sample.addStack(stack)
    sample.addLayer(ba.Layer(si_mat))
    return sample


def simulate(sample):
    n = 500
    lam = 1.54*angstrom

    scan_a = ba.AlphaScan(n, 5*deg/n, 5*deg)
    scan_a.setWavelength(lam)
    scan_a.setIntensity(INTENSITY)
    result_a = ba.SpecularSimulation(scan_a, sample).simulate()
    result_a.setTitle("angle scan")

    qs = [4*pi*sin(a)/lam
          for a in np.linspace(5*deg/n, 5*deg, n)]
    scan_q = ba.QzScan(qs)
    scan_q.setIntensity(INTENSITY)
    result_q = ba.SpecularSimulation(scan_q, sample).simulate()
    result_q.setTitle("Q scan")

    return [result_a, result_q]


if __name__ == '__main__':
    sample = get_sample()
    results = simulate(sample)
    plotargs = bp.parse_commandline()
    bp.plot_multicurve(results, **plotargs)
    bp.export(**plotargs)
